<?php

class ZEUSLogGroup extends LOGGROUP {
 var $table;
 
 function __construct(array &$info, ZEUS $zeus, $flags = 0) {
    parent::__construct($info);
    
    if ($this->gid === false) $this->gid = 0;
    $this->table = "log" . $this->gid;
 }
}

class ZEUSMask extends MASK {
 function __construct(&$props = NULL, ZEUS $zeus) {
    if (($props)&&(preg_match("/^maskid(\d+)$/",$props["db_mask"], $m))) {

	$sql = "SELECT name, mask FROM masks WHERE maskid=" . $m[1];
	try {
	    $stmt = $zeus->db->Prepare($sql);
	    $stmt->bindColumn(1, $name);
	    $stmt->bindColumn(2, $mask);//, PDO::PARAM_LOB);
	    $stmt->execute();

	    $row = $stmt->fetch(PDO::FETCH_BOUND);
	    unset($stmt);
	    
	    if ($row) {
		$zeus_mask = $zeus->ConvertBinary($mask, ZEUS::TYPE_INT32);
		$this->ids = array();
		for ($i=0;isset($zeus_mask[$i]);$i++) {
		    if ($zeus_mask[$i] < 2) {
			array_push($this->ids, $i);
		    }
		}
		$this->name = $name;
		return;
	    }
	} catch (PDOException $pe) {
	    $e = $this->dbh->errorInfo();
	    throw new ADEIException(translate("Execution of the SQL Query is failed. SQL Error: %u, Driver Error: %u, Message: %s ", $e[0], $e[1], $e[2]) . "[$sql]");
	}
	
    }
    
    parent::__construct($props);
 }
}

class ZEUSData implements Iterator {
 var $zeus;
 var $stmt;
 var $row;
 
 function __construct(ZEUS $zeus, PDOStatement $stmt) {
    $this->zeus = $zeus;
    $this->stmt = $stmt;
 }
 
 function rewind() {
    try {
	$this->stmt->execute();
	$this->next();
    } catch(PDOException $e) {
	throw new ADEIException(translate("SQL request is failed with error") . ": " . $e->getMessage(), $e->getCode());
    }

 }
 
    // current element
 function current() {
    if ($this->zeus->char_mode) {
	$data = pack("H*", $this->row["data"]);
    } else {
	$data = &$this->row["data"];
    }
    
    $kanz=(strlen($data)-4)/8;
    $bform='d'.$kanz.'ist/Vlen';
    $blob = unpack($bform,strrev($data));

    $res = array();
    for ($i=0;$i<$kanz;$i++) {
	if (($kanz > 1)||(isset($blob["ist".($kanz-$i)]))) 
	    $res[$i]=$blob["ist".($kanz-$i)];
	else 
	    $res[$i]=$blob["ist"];
    }
    return $res;
 }
 
    // current key (PHP rounding to integer :()
 function key() {
    return $this->zeus->ExportUnixTime($this->row['ts']);
 }
 
    // advvance to next (and returns it or false)
 function next() {
    try {
//	do {
	    $this->row = $this->stmt->fetch();
//	    echo "_________________________ " . $this->row['ts'] . " (" . strlen($this->row['data']) . ", " . ((strlen($this->row['data'])-4)/8) . ")\n";
//	} while (($this->row)&&(strlen($this->row["data"]) < 12));
    } catch(PDOException $e) {
	throw new ADEIException(translate("SQL error") . ": " . $e->getMessage());
    }
 }
 
    // checks if there is current element
 function valid() {
    return $this->row?true:false;
 }
}

class ZEUS extends READER {
 var $db;
 var $char_mode; /* Specifies whatever SQL_C_CHAR or SQL_C_BINARY is returned 
	by database while BLOB is requested */
			
 const TYPE_FLOAT = 1;
 const TYPE_INT32 = 2;

 function __construct(&$props) {
    parent::__construct($props);    
    $this->db = new DATABASE($this->server);

    switch ($this->server["driver"]) {
	case "odbc":
	    $this->char_mode = true;
	    break;
	default:
	    $this->char_mode = false;
    }
 }
 
 function ConvertBinary(&$zeus_blob, $type = ZEUS::TYPE_FLOAT) {
    if ($this->char_mode) {
	$data = pack("H*", $zeus_blob);
    } else {
	$data = &$zeus_blob;
    }
    
    switch ($type) {
	case ZEUS::TYPE_FLOAT:
	    $kanz=(strlen($data)-4)/8;
	    $bform='d'.$kanz.'ist/Vlen';
	    break;
	case ZEUS::TYPE_INT32:
	    $kanz=(strlen($data)-4)/4;
	    $bform='V'.$kanz.'ist/Vlen';
	    break;
	default:
	    throw new ADEIException(translate("Unknown ZEUS type (%i) is specified", $type));
    }	
    $blob = unpack($bform,strrev($data));

    $res = array();
    for ($i=0;$i<$kanz;$i++) {
	if (($kanz > 1)||(isset($blob["ist".($kanz-$i)]))) 
	    $res[$i]=$blob["ist".($kanz-$i)];
	else 
	    $res[$i]=$blob["ist"];
    }
    return $res;
 }
    
 function CreateGroup(array &$ginfo = NULL, $flags = 0) {
    if (!$ginfo) $ginfo = $this->req->GetGroupInfo();
    return new ZEUSLogGroup($ginfo, $this);
 }
 
 function CheckGroup(LOGGROUP &$grp = NULL) {
    if ($grp) {
	if (!$grp instanceof ZEUSLogGroup)
	    throw Exception(translate("Invalid LOGGROUP supplied"));
    } else {
	$grinfo = $this->req->GetGroupInfo();
	$grp = $this->CreateGroup($grinfo);
    }
    return $grp;
 }
 
 function CreateMask(LOGGROUP &$grp = NULL, array &$minfo = NULL, $flags = 0) {
    if (!$minfo) {
	if ($this->req instanceof GROUPRequest)
	    $minfo = $this->req->GetMaskInfo();
    }
    
    return new ZEUSMask($minfo, $this);
 }
 
/* ToDo: 
 *  1. Add checks (if it is really Zeus database)
 */
 function GetDatabaseList($flags = 0) {
    global $READER_DB;


    if (!($flags&REQUEST::LIST_WILDCARDED)) return $this->db->GetDatabaseList();
    
    if ($this->req)
	$server = $this->req->GetServerConfig();
    else
	throw new ADEIException(translate("The data source server is not specified"));


    $filter = "";
    foreach ($server['database'] as $db) {
	if ($filter) $filter .= "|$db";
	else $filter = $db;
    }
    if ($server['multibase']) {
	if (is_string($server['multibase'])) {
	    $filter = "/^(" . $filter . ")" . $server['multibase'] . "/";
	} else {
	    $filter = "/^(" . $filter . ")/";
	}
    } else {
	$filter = "/^(" . $filter . ")$/";
    }
    
    return array_reverse($this->db->GetDatabaseList($filter));
 }
 
 function GetGroupInfo(LOGGROUP &$grp = NULL, $flags = 0) {
    if ($flags&REQUEST::NEED_INFO)
	$req_cols = "*";
    else
	$req_cols = "gid, name";

    if ($grp)
	$res = $this->db->Query("SELECT $req_cols FROM groups WHERE gid=" . $grp->gid);
    else
	$res = $this->db->Query("SELECT $req_cols FROM groups");

    $groups = array();
    if ($flags&REQUEST::NEED_INFO) {
	foreach ($res as $row) {
	    $gid = $row['gid'];
	    $groups[$gid] = $row;
        
	    $ginfo = array("db_group" => $gid);
	    $grzeus = $this->CreateGroup($ginfo);
	    
	    $req = "MIN(ts), MAX(ts)";
	    if ($flags&REQUEST::NEED_COUNT) 
		$req .= ", COUNT(ts)";

	    $valres = $this->db->Query("SELECT $req FROM " . $grzeus->table, DATABASE::FETCH_NUM);
	    $vals = $valres->fetch(PDO::FETCH_NUM);
	    $valres = NULL;
	    
	    $groups[$gid]['__internal__'] = array();

	    $groups[$gid]['first'] = $this->ExportUnixTime($vals[0]);
	    if (($this->start_date)&&($this->start_date > $groups[$gid]['first'])) {
	    	$groups[$gid]['first'] = $this->start_date;
		$groups[$gid]['__internal__']['first'] = $this->ImportUnixTime($this->start_date);
	    } else {
	        $groups[$gid]['__internal__']['first'] = $vals[0];
	    }

	    $groups[$gid]['last'] = $this->ExportUnixTime($vals[1]);
	    if (($this->end_date)&&($this->end_date < $groups[$gid]['last'])) {
	    	$groups[$gid]['last'] = $this->end_date;
		$groups[$gid]['__internal__']['last'] = $this->ImportUnixTime($this->end_date);
	    } else {
		$groups[$gid]['__internal__']['last'] = $vals[1];
	    }

	    if ($flags&REQUEST::NEED_COUNT) 
		$groups[$gid]['records'] = $vals[2];

	    if ($flags&REQUEST::NEED_ITEMINFO) {
		$groups[$gid]['items'] = $this->GetItemList($grzeus);
	    }

	}
    } else {
	foreach ($res as $row) {
	    $groups[$row['gid']] = $row;
	}
    }

    return $grp?$groups[$grp->gid]:$groups;
 }
 
 function GetExperimentList($flags = 0) {
    $resp = $this->db->query("SELECT experiment, name, start, stop FROM marker");

    $list = array();
    foreach ($resp as $row) {
	if (!preg_match("/[\w\d]/", $row['name'])) $row['name'] = _("No name");
	    /* Read format */
	    
	$item = array();
	$item['start'] = $row['start'];
	$item['stop'] = $row['stop'];
	    
	if (preg_match("/[\w\d]/", $row['experiment'])) $have_exp = 1;
	else $have_exp = 0;
	if (preg_match("/[\w\d]/", $row['name'])) $have_name = 1;
	else $have_name = 0;
	    
	    
	if (($have_exp)&&($have_name)) $item['name'] = $row['experiment'] . ": " . $row['name'];
	else if ($have_exp) $item['name'] = $row['experiment'];
	else if ($have_name) $item['name'] = $row['name'];
	else $item['name'] = $item['start'] . " - " . $item['stop'];
	    
	array_push($list, $item);
    }

    return $list;
 }

 function GetMaskList(LOGGROUP &$grp = NULL, $flags = 0) {
    $grp = $this->CheckGroup($grp);

    $list = array();
    
    $resp = $this->db->Query("SELECT maskid, name FROM masks WHERE gid=" . $grp->gid);
    foreach ($resp as $row) {
	if (!preg_match("/[\w\d]/", $row['name'])) $row['name'] = _("No name");
	$list["maskid" . $row['maskid']] = $row;
    }

    return $list;
 }
 
 function GetItemList(LOGGROUP &$grp = NULL, MASK &$mask = NULL, $flags = 0) {
    $grp = $this->CheckGroup($grp);
    if (!$mask) $mask = $this->CreateMask($grp);

    $bids = array();
    $items = array();
    
    $resp = $this->db->Query("SELECT id FROM g2id WHERE gid=" . $grp->gid . " ORDER BY pos ASC");
    foreach ($resp as $row) {
	array_push($bids, $row['id']);
    }

    $pos = 0; $rpos = 0;
    foreach ($bids as $bid) {    

	$resp = $this->db->Query("SELECT length, name, itemnames FROM blocks WHERE bid=$bid");
	if ($resp) $res = $resp->fetch();
	else $res = false;
	    
	$names = preg_split("/\r?\n/", $res['itemnames']);
	for ($i = 0; $i < $res['length']; $i++, $pos++) {
	    if (!$mask->Check($pos)) continue;
	    
	    $items[$rpos] = array(
		"id" => $pos,
		"group" => $res['name'],
		"name" =>  $names[$i],
		"opc" => ""
	    );
	    if (!preg_match("/[\w\d]/", $items[$rpos]["name"])) {
		if (preg_match("/[\w\d]/", $items[$rpos]["group"])) {
		    $items[$rpos]["name"] = "item" . $i;
		} else {
		    $items[$rpos]["name"] = "item" . $pos;
		}
	    }
	    if (preg_match("/[\w\d]/", $items[$rpos]["group"])) {
		$items[$rpos]["name"] = $items[$rpos]["name"] . " [" . $items[$rpos]["group"] . "]";
	    } 
	    
	    $rpos++;
	}
	unset($names);
    }
    
    return $items;
 }

    /* ZEUS time is a LabVIEW time */
 function ImportUnixTime($unix_time) {
    if (!$unix_time) return 0;
    return ($unix_time + (2082837600+7200));
 }
 
 function ExportUnixTime($zeus_time) {
    if (!$zeus_time) return false;
    return ($zeus_time - (2082837600+7200));

 }
 
 function GetGroupSize(LOGGROUP &$grp = NULL) {
    $grp = $this->CheckGroup($grp);

    $bids = array();
    
    try {    
	$resp = $this->db->Query("SELECT id FROM g2id WHERE gid=" . $grp->gid);
	foreach ($resp as $row) {
	    array_push($bids, $row['id']);
	}

	$size = 0;
	foreach ($bids as $bid) {    
	    $resp = $this->db->Query("SELECT length FROM blocks WHERE bid=$bid");
	    if ($resp) $res = $resp->fetch();
	    else $res = false;
	    if ($res) $size += $res['length'];
	}
    } catch (PDOException $e) {
	throw new Exception($e->getMessage());
    }
    
    return $size;
 }

    /* Between includes start point, but excludes end point */ 
 function GetData(LOGGROUP &$grp = NULL, $from = 0, $to = 0) {
    $grp = $this->CheckGroup($grp);

    if ((!$from)||(!$to)) {
        $ivl = $this->CreateInterval($grp);
	$ivl->Limit($from, $to);
	
	$from = $ivl->window_start;
	$to = dsMathPreciseAdd($ivl->window_start, $ivl->window_size);
    }

    $stmt = $this->db->Prepare("SELECT ts, data FROM " . $grp->table . " WHERE ts BETWEEN " . $this->ImportUnixTime($from) . " and " . $this->ImportUnixTime($to));
    return new ZEUSData($this, $stmt);
 }
 
 function RemoveData(LOGGROUP &$grp, $time, $to = 0) {
    if ($to) {
	$this->db->Query("DELETE FROM " . $grp->table . " WHERE ts BETWEEN " . $this->ImportUnixTime($time) . " and " . $this->ImportUnixTime($to));
    } else {
	$dbtime = floor($this->ImportUnixTime($time));
	$this->db->Query("DELETE FROM " . $grp->table . " WHERE ts BETWEEN " . $dbtime . " and " . ($dbtime + 1));
    }
 }

 function PushData(LOGGROUP &$grp, $time, $data) {
    $dbtime = $this->ImportUnixTime($time);
    $size = sizeof($data['values']);
    
    $res = "";
    for ($i=$size;$i>0;$i--) $res .= pack("d", $data['values'][$i-1]);
    $res .= pack("V", $size);
    $blob = strrev($res);

/*    
    $kanz=(strlen($blob)-4)/8;
    $bform='d'.$kanz.'ist/Vlen';
    $blob = unpack($bform,strrev($blob));

    $res = array();
    for ($i=0;$i<$kanz;$i++) {
	if (($kanz > 1)||(isset($blob["ist".($kanz-$i)]))) 
	    $res[$i]=$blob["ist".($kanz-$i)];
	else 
	    $res[$i]=$blob["ist"];
    }

//    print date("c", $time) . ".  " . $res[7] . " = " . $data['values'][7] . "\n";
    
*/

    $dst = $this->db->Prepare("INSERT INTO " . $grp->table . " (ts,data) VALUES(?,?)");
    $dst->bindParam(1, $dbtime);
    $dst->bindParam(2, $blob, PDO::PARAM_LOB);
    $dst->execute();

 }
 
 function Backup(&$binfo, $flags = 0) {
     global $BACKUP_DB;

    if (is_array($dbinfo)) {
	$dbinfo = &$binfo;
    } else {
	$dbinfo = $BACKUP_DB;
	$dbinfo['database'] = $binfo;
    }	
    
    try {
	$zeus  = new ZEUS($dbinfo);
    } catch (ADEIException $e) {
	if ($flags&READER::BACKUP_FULL) {
	    $tmpinfo = $dbinfo;
	    unset($tmpinfo['database']);

	    $zeus = new ZEUS($tmpinfo);
	    $dblist = $zeus->db->ShowDatabases();
	    
	    foreach ($dblist as $row) {
		if (strtolower($row[0]) == strtolower($dbinfo['database'])) {
		    throw new ADEIException(translate("Broken or invalid database is specified for backup"));
		}
	    }

	    if ((strtolower($dbinfo['driver']) == "mysql")&&(strtolower($this->server['driver']) == "mysql")) {
		$cli = Database::GetConnectionString($this->server);
		$srv = Database::GetConnectionString($dbinfo);

		$zeus->db->CreateDatabase($dbinfo['database']);
		@system("mysqldump --compact $cli | mysql $srv", $res);
		if ($res) throw new ADEIException(translate("The initialization of the backup database is failed."));
		
		$zeus = new ZEUS($dbinfo);
	    } else throw new ADEIException(translate("The initial backup database should setup manualy. The auto mode is only works for the MySQL at the moment"));
	} else throw $e;
    }


    $src_list = $this->GetGroupList(REQUEST::NEED_INFO|REQUEST::NEED_ITEMINFO);
    $dst_list = $zeus->GetGroupList(REQUEST::NEED_INFO|REQUEST::NEED_ITEMINFO);
    
    if ((sizeof($src_list) != sizeof($dst_list))||(sizeof($src_list['items'])!=sizeof($dst_list['items']))) {
	// or compare names (items by content not lenght, 'name', 'comment')
	throw new ADEIException(translate("The configuration of the current backup not in sync with the data source"));
    }
	    
    
    $list = $this->req->GetGroups();
    foreach ($list as $greq) {
	$grp = $greq->CreateGroup($this);
	
	if ($dst_list[$grp->gid]['__internal__']['last']) $from = $dst_list[$grp->gid]['__internal__']['last'];
	else $from = 0;

	if ($src_list[$grp->gid]['__internal__']['last']) $to = $src_list[$grp->gid]['__internal__']['last'];
	else $to = 0;

	if (abs(round($from) - round($to))<2) continue;

	if ($from > $to) 
	    throw new ADEIException(translate("The backup had newer data than the source does"));

	try {
	    $src = $this->db->Prepare("SELECT lid,ts,data FROM " . $grp->table . " WHERE ts > " . $from . " ORDER by lid ASC");
	    if (!$src->execute()) {
		$einfo = $dst->errorInfo();
		throw new ADEIException("Failure while backing up group " . $grp->gid . ". SQL Error: " . $einfo[0] . ", Driver Error: " . $einfo[1] . ", Message: " . $einfo[2]);
	    }
	    $src->bindColumn(1, $lid);
	    $src->bindColumn(2, $ts);
	    $src->bindColumn(3, $data, PDO::PARAM_LOB);
	    
	    $dst = $zeus->db->Prepare("INSERT INTO " . $grp->table . " (lid,ts,data) VALUES (?, ?, ?)");
	    $dst->bindParam(1, $lid);
	    $dst->bindParam(2, $ts);
	    $dst->bindParam(3, $data, PDO::PARAM_LOB);
	
	    while ($src->fetch(PDO::FETCH_BOUND)) {
		if (!$dst->execute()) {
		    $einfo = $dst->errorInfo();
		    throw new ADEIException("Failure while backing up group " . $grp->gid . ", item " . $lid . "(" . date("r", ceil($this->ExportUnixTime($ts))) . ") SQL Error: " . $einfo[0] . ", Driver Error: " . $einfo[1] . ", Message: " . $einfo[2]);
		}
	    }
	} catch(PDOException $e) {
	    throw new ADEIException(translate("SQL request is failed with error") . ": " . $e->getMessage(), $e->getCode());
	}
    }
 }
}

?>